# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2007, 2008, 2009 Arnaud Ebalard
#               2015, 2016, 2017 Maxence Tury

"""
Aggregate some TLS crypto objects.
"""

from scapy.layers.tls.crypto.suites import *  # noqa: F401
